try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore
try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""
A few random plotting functions.
"""

import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import celerite
from eztao.carma.CARMATerm import *
from eztao.ts.carma_sim import pred_lc


def plot_drw_ll(
    t,
    y,
    yerr,
    best_params,
    gp,
    prob_func,
    amp_range=None,
    tau_range=None,
    nLevels=10,
    **kwargs,
):
    """
    Plot DRW log likelihood surface.

    Args:
        t (array(float)): Time stamps of the input time series (the default unit is day).
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        best_params (array(float)): Best-fit DRW parameters, [amp, tau].
        gp (object): celerite GP object with a proper DRW kernel.
        prob_func (func): Posterior/Likelihood function with args=(params, y, gp)
        amp_range (tuple, optional): The range of parameters to evaluate likelihood.
            Defaults to None.
        tau_range (tuple, optional): The range of parameters to evaluate likelihood.
            Defaults to None.
        nLevels (int, optional): The number of levels in the final contour plot.
            Defaults to 10.

    Keyword Args:
        grid_size (int): The number of points to evaluate likelihood along a given axis.
        true_params (array(float)): The true DRW parameters of the input time series.
    """
    best_amp, best_tau = best_params
    grid_size = 40

    if amp_range is None:
        amp_range = [best_amp / np.e, np.e * best_amp]
    if tau_range is None:
        tau_range = [best_tau / np.e, np.e * best_tau]

    # check if custom resolution
    if "grid_size" in kwargs:
        grid_size = int(kwargs.get("grid_size"))

    log_taus = np.linspace(np.log(tau_range[0]), np.log(tau_range[1]), grid_size)
    log_amps = np.linspace(np.log(amp_range[0]), np.log(amp_range[1]), grid_size)

    param_grid = np.meshgrid(log_amps, log_taus)  # log scale
    # reg_param_grid = np.exp(param_grid)  # normal scale

    # flatten to list
    ls_params_combo = list(zip(param_grid[0].flatten(), param_grid[1].flatten()))

    # run in bulk
    gp.compute(t, yerr)
    vec_neg_ll = np.vectorize(prob_func, excluded=[1, 2], signature="(n)->()")
    bulk_ll = np.negative(vec_neg_ll(ls_params_combo, y, gp))
    ll_reshape = bulk_ll.reshape((grid_size, grid_size))

    # normalize, better contours
    max_ll, min_ll = np.max(bulk_ll), np.min(bulk_ll)
    ll_range = max_ll - min_ll
    divnorm = mpl.colors.TwoSlopeNorm(
        vmin=min_ll, vmax=max_ll + 1, vcenter=(5 + max_ll - 0.02 * ll_range)
    )

    # compute contour levels
    delta_levels = np.exp(np.linspace(0, np.log(ll_range), 10, endpoint=False))
    levels = max_ll - delta_levels
    levels = levels[::-1]

    # plot the main axes
    fig = plt.figure(dpi=150)
    ax = fig.add_axes([0.1, 0.1, 0.6, 0.8])
    img = plt.contourf(
        param_grid[0],
        param_grid[1],
        ll_reshape,
        levels,
        norm=divnorm,
        extend="both",
        alpha=0.9,
    )
    plt.contour(
        param_grid[0], param_grid[1], ll_reshape, levels, colors=("k",), linewidths=0.5
    )
    ax.yaxis.set_major_locator(plt.MaxNLocator(7, prune="both"))
    ax.xaxis.set_major_locator(plt.MaxNLocator(5))

    # colorbar
    cbar_ax = fig.add_axes([0.74, 0.1, 0.03, 0.8])
    cbar = plt.colorbar(img, cax=cbar_ax, format="%.f")
    cbar.minorticks_off()

    ax.scatter(
        np.log(best_params[0]),
        np.log(best_params[1]),
        marker="*",
        s=150,
        label="Best-fit",
        zorder=10,
    )

    if "true_params" in kwargs:
        true_params = kwargs.get("true_params")
        ax.scatter(
            np.log(true_params[0]),
            np.log(true_params[1]),
            marker="*",
            s=150,
            label="True",
            zorder=10,
        )

    ax.set_title("Loglikelihood Surface")
    ax.set_xlabel(r"$Log$(Amplitude)")
    ax.set_ylabel(r"$Log\/(\tau)$")
    ax.legend(markerscale=1)


def plot_dho_ll(
    t,
    y,
    yerr,
    best_params,
    gp,
    prob_func,
    inner_dim=10,
    outer_dim=4,
    ranges=[(None, None), (None, None), (None, None), (None, None)],
    nLevels=10,
    **kwargs,
):
    """
    Plot DHO/CARMA(2,1) log likelihood lanscape.

    Args:
        t (array(float)): Time stamps of the input time series (the default unit is day).
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        best_params (array(float)): Best-fit DHO parameters in [a1, a2, b0, b1].
        gp (object): celerite GP object with a proper DHO kernel.
        prob_func (func): Posterior/Likelihood function with args=(params, y, gp).
        inner_dim (int, optional): The number of points to eval likelihood along
            a1 and a2. Defaults to 10.
        outer_dim (int, optional): The number of points to eval likelihood along
            b0 and b1. Defaults to 4.
        ranges (list, optional): Parameters ranges (in natural log) within which to plot
            the surface. Defaults to [(None, None), (None, None), (None, None),
            (None, None)].
        nLevels (int, optional): The number of levels in the final contour plot. 
            Defaults to 10.

    Keyword Args:
        true_params (array(float)): The true DHO parameters of the input time series.
    """
    best_log = np.log(best_params)
    num_param = len(best_log)
    param_range = np.zeros((num_param, 2))

    # check if any ranges provided
    for i in range(num_param):
        if any(map(lambda x: x is None, ranges[i])):
            param_range[i, :] = np.array([best_log[i] - 1, best_log[i] + 1])
        else:
            param_range[i, :] = np.array(ranges[i])

    # create meshgrid to compute ll
    inner_grid_params = [0, 1]
    outer_grid_params = [2, 3]
    grid_ls = []

    for i in range(num_param):
        pm_range = param_range[i]

        # check if param is in inner grid
        if i in inner_grid_params:
            grid_ls.append(np.linspace(pm_range[0], pm_range[1], inner_dim))
        else:
            wide_grid = np.linspace(pm_range[0], pm_range[1], outer_dim + 1)
            delta = wide_grid[1] - wide_grid[0]
            grid_ls.append(wide_grid[1:] - delta / 2)

    param_grid = np.meshgrid(*grid_ls)  # log scale

    # prepare for np.vectorize
    vec_neg_ll = np.vectorize(prob_func, excluded=[1, 2], signature="(n)->()")
    param_combos = zip(
        param_grid[0].flatten(),
        param_grid[1].flatten(),
        param_grid[2].flatten(),
        param_grid[3].flatten(),
    )
    ls_param_combos = list(param_combos)

    # run in bulk
    gp.set_parameter_vector(best_log)
    gp.compute(t, yerr)
    bulk_ll = np.negative(vec_neg_ll(ls_param_combos, y, gp))

    # back to shape (dim, dim, dim, dim)
    dims = np.empty(num_param, dtype=np.int)
    dims[inner_grid_params] = inner_dim
    dims[outer_grid_params] = outer_dim
    ll_reshape = bulk_ll.reshape(tuple(dims))

    # normalize & compute contour levels
    max_ll, min_ll = np.max(bulk_ll), np.min(bulk_ll)
    ll_range = max_ll - min_ll
    delta_levels = np.exp(np.linspace(0, np.log(ll_range), 10, endpoint=False))
    levels = max_ll - delta_levels
    levels = levels[::-1]
    divnorm = mpl.colors.TwoSlopeNorm(
        vmin=min_ll, vmax=max_ll + 1, vcenter=(5 + max_ll - 0.1 * ll_range)
    )

    # determine the frame containing the best ll
    idx_max = np.unravel_index(
        np.argmax(np.max(ll_reshape, axis=(0, 1)), axis=None), (outer_dim, outer_dim)
    )

    # plot
    fig, axs = plt.subplots(
        outer_dim, outer_dim, figsize=(9, 8), sharey=True, sharex=True, dpi=200
    )
    images = {}
    b0s = grid_ls[2]
    b1s = grid_ls[3]

    # loop over each grid point on outer layer
    for i in range(outer_dim):
        for j in range(outer_dim):
            _ = axs[i, j].contourf(
                param_grid[0][:, :, i, j],
                param_grid[1][:, :, i, j],
                ll_reshape[:, :, i, j],
                levels,
                extend="both",
            )
            images[f"{i}_{j}"] = _
            axs[i, j].scatter(
                best_log[0], best_log[1], marker="*", s=100, c="b", alpha=0.8
            )
            if "true_params" in kwargs:
                true_params = kwargs.get("true_params")
                axs[i, j].scatter(
                    true_params[0],
                    true_params[1],
                    marker="*",
                    s=100,
                    label="True",
                    c="orange",
                )
            axs[i, j].text(
                0.1,
                0.8,
                f"log(b0):{b0s[i]:.2f}\nlog(b1):{b1s[j]:.2f}",
                fontdict={"size": 6, "color": "w", "weight": 550},
                transform=axs[i, j].transAxes,
            )

    # reset the norm
    plt.subplots_adjust(hspace=0, wspace=0)
    for key in images:
        images[key].set_norm(divnorm)

    fig.colorbar(
        images[f"{idx_max[0]}_{idx_max[1]}"], ax=axs, fraction=0.05, format="%.0f"
    )

    # title and axis label
    fig.text(0.46, 0.04, "log(a1)", ha="center", fontsize=15)
    fig.text(0.04, 0.5, "log(a2)", va="center", rotation="vertical", fontsize=15)
    axs[idx_max[0], idx_max[1]].scatter(
        best_log[0], best_log[1], marker="*", s=150, c="r"
    )
    fig.suptitle("DHO Loglikelihood Surface", x=0.5, y=0.92)


def plot_pred_lc(t, y, yerr, params, p, t_pred, plot_input=True):
    """
    Plot GP predicted time series given best-fit parameters.

    Args:
        t (array(float)): Time stamps of the input time series.
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        params (array(float)): Best-fit CARMA parameters
        p (int): The AR order (p) of the best-fit model.
        t_pred (array(float)): Time stamps at which to generate predictions.
        plot_input (bool): Whether to plot the input time series. Defaults to True.
    """

    # get pred lc
    t_pred, mu, var = pred_lc(t, y, yerr, params, int(p), t_pred)

    # Plot the data
    t_range = t_pred[-1] - t_pred[0]
    fig, ax = plt.subplots(1, 1, dpi=150, figsize=(8, 4))
    color = "#ff7f0e"
    ax.plot(t_pred, mu, color=color, label="Mean Prediction", alpha=0.8)

    # if valid variance returned
    if np.median(var) > (np.median(np.abs(yerr)) / 1e10):
        std = np.sqrt(var)
        ax.fill_between(
            t_pred, mu + std, mu - std, color=color, alpha=0.3, edgecolor="none"
        )

    if plot_input:
        ax.errorbar(
            t, y, yerr=yerr, fmt=".k", capsize=2, label="Input Data", markersize=3
        )

    # other plot setting
    ax.set_xlim(t_pred[0] - t_range / 50, t_pred[-1] + t_range / 50)
    ax.set_ylabel(r"Flux (arb. unit)")
    ax.set_xlabel(r"Time (day)")
    ax.yaxis.set_major_locator(plt.MaxNLocator(5))
    ax.set_title("Maximum Likelihood Prediction")
    ax.legend()
    fig.tight_layout()



"""
A collection of utility functions to assist analysis/simulation of time series data.
"""

import numpy as np
from numba import njit
from functools import partial
from scipy.stats import median_abs_deviation as mad

__all__ = ["downsample_byN", "add_season", "downsample_byTime", "median_clip"]


def downsample_byN(t, nObs):
    """
    Randomly choose N data points from a given time series.

    Args:
        t (array(float)): Time stamps of the original time series.
        N (int): The number of entries in the final time series.

    Returns:
        A 1d array booleans indicating which data points to keep.
    """
    # random choose index
    idx = np.arange(len(t))
    mask = np.zeros_like(idx, dtype=np.bool_)
    true_idx = np.random.choice(idx, nObs, replace=False)

    # assign chosen index to 1/True
    mask[true_idx] = 1

    return mask


def _get_nearest_idx(tIn, x):
    """Internal function to return nearest neighbor in an array."""
    return (np.abs(tIn - x)).argmin()


def downsample_byTime(tIn, tOut):
    """
    Downsample a time series at desired output time stamps.

    Args:
        tIn (array(float)): Time stamps of the original time series.
        tOut (array(float)): Time stamps of the output time series.

    Returns:
        array(int): Indices for which the data points should be kept from the original
        time series. Note that there could be duplicates.
    """
    get_nearest = partial(_get_nearest_idx, tIn)
    return np.array(list(map(get_nearest, tOut)))


def add_season(t, lc_start=0, season_start=90, season_end=270):
    """
    Insert seasonal gaps into time series

    Args:
        t (array(float)): Time stamps of the original time series.
        lc_start (float): Starting day for the output time series. (0 -> 365.25).
            Default to 0.
        season_start (float): Observing season start day within a year. Default to 90.
        season_end (float): Observing season end day within a year. Default to 270.

    Returns:
        A 1d array booleans indicating which data points to keep.
    """
    t = t - t[0]
    t = t + lc_start

    mask = (np.mod(t, 365.25) > season_start) & (np.mod(t, 365.25) < season_end)

    return mask


def median_clip(y, num_sigma=3):
    """Clip time series using a three point median filter.

    The sigma (standard deviation) for the time series is computed from the median  absolute deviation (MAD) as to reduce the effects from extreme outliers, where
    sigma \sim 1.4826*MAD. If more than 10% of the data points are removed, the upper
    bound will be lifted gradually until that fraction drops bellow 10%.

    Args:
        y (array(float)): y values of the original time series.
        num_sigma (int, optional): Data points that are more than this number of sigma
            away from the three point median will be removed. Defaults to 3.

    Returns:
        A 1d array booleans indicating which data points to keep.
    """
    y = np.atleast_1d(y)
    lc_len = y.shape[0]
    med = y.copy()
    med[1:-2] = [np.median(y[i - 1 : i + 2]) for i in range(1, lc_len - 2)]

    # need to deal with edges of median filter
    med[0] = np.median(np.array([y[-1], y[0], y[1]]))
    med[-1] = np.median(np.array([y[-2], y[-1], y[0]]))

    # compute residual & sigma
    res = np.abs(y - med)
    sigma = mad(res, scale='normal')

    # set clipping thresh hold
    raise_bar = True
    thresh = num_sigma * sigma

    # if remove too much, raise bar until only remove 10%
    while raise_bar:
        ratio = np.sum(res > thresh) / lc_len
        if ratio < 0.1:
            break
        else:
            thresh += 0.1 * sigma
    return res < thresh



"""
A collection of functions to fit/analyze time series using CARMA models.
"""

import numpy as np
from math import ceil
from scipy.optimize import minimize
from scipy.stats import median_abs_deviation as mad
import celerite
from celerite import GP
from eztao.carma.CARMATerm import DRW_term, DHO_term, CARMA_term
from functools import partial

# change numpy warn setting
old_settings = np.seterr(all="warn")
np.seterr(invalid="ignore")

__all__ = [
    "drw_fit",
    "dho_fit",
    "carma_fit",
    "flat_prior",
    "neg_fcoeff_ll",
    "neg_param_ll",
    "neg_lp_flat",
    "drw_log_param_init",
    "dho_log_param_init",
    "carma_log_fcoeff_init",
    "sample_carma",
    "scipy_opt",
]


def neg_fcoeff_ll(log_fcoeffs, y, gp):
    """
    Negative log likelihood function for CARMA specified in the factored poly space.

    This method will catch 'overflow/underflow' runtimeWarning and
    return inf as probability.

    Args:
        log_fcoeffs (array(float)): Coefficients (in natural log) of a CARMA model in
            the factored polynomial space.
        y (array(float)): y values of the input time series.
        gp (object): celerite GP object with a proper CARMA kernel.

    Returns:
        float: Negative log likelihood.
    """

    assert gp.kernel.p >= 2, "Use neg_param_ll() instead!"
    neg_ll = np.inf

    try:
        gp.kernel.set_log_fcoeffs(log_fcoeffs)
        neg_ll = -gp.log_likelihood(y)
    except celerite.solver.LinAlgError as c:
        # print(c)
        pass
    except Exception as e:
        pass

    return neg_ll


def neg_param_ll(log_params, y, gp):
    """
    Negative log likelihood function for CARMA specified in the nominal space.

    This method will catch 'overflow/underflow' runtimeWarning and
    return inf as probability.

    Args:
        log_params (array(float)): Natural log of CARMA parameters.
        y (array(float)): y values of the input time series.
        gp (object): celerite GP object with a proper CARMA kernel.

    Returns:
        float: Negative log likelihood.
    """
    assert gp.kernel.p <= 2, "Use neg_fcoeff_ll() instead!"
    neg_ll = np.inf

    try:
        gp.set_parameter_vector(log_params)
        neg_ll = -gp.log_likelihood(y)
    except celerite.solver.LinAlgError as c:
        # print(c)
        pass
    except Exception as e:
        pass

    return neg_ll


def flat_prior(log_params, bounds):
    """
    A flat prior function. Returns 0 if "log_params" are within the given "bounds",
    negative infinity otherwise.

    Args:
        log_params (array(float)): CARMA parameters in natural log.
        bounds (array((float, float)): An array of boundaries.

    Returns:
        float: 0 or negative infinity.
    """

    dim = len(log_params)
    assert bounds.shape == (dim, 2), "Dimension mismatch for the boundaries!"

    bounds_idx = [i for i in range(dim) if all(bounds[i])]
    in_bounds = np.array(
        [(bounds[i, 0] <= log_params[i] <= bounds[i, 1]) for i in bounds_idx]
    )

    if all(in_bounds):
        return 0
    else:
        return -np.inf


def neg_lp_flat(log_params, y, gp, bounds=None, mode="fcoeff"):
    """
    Negative log probability function using a flat prior.

    Args:
        log_params (array(float)): CARMA parameters (or coefficients of the factored
            characteristic polynomial) in natural log.
        y (array(float)): y values of the input time series.
        gp (object): celerite GP object with a proper CARMA kernel.
        bounds (array((float, float)): An array of boundaries. Defaults to None.
        mode (str, optional): The parameter space in which proposals are made. The mode
            determines which loglikehood function to use. Defaults to "fcoeff".

    Returns:
        float: Log probability of the proposed parameters.
    """

    if mode == "param":
        neg_ll = neg_param_ll
    else:
        neg_ll = neg_fcoeff_ll

    return -flat_prior(log_params, bounds) + neg_ll(log_params, y, gp)


def drw_log_param_init(amp_range, log_tau_range, size=1):
    """
    Randomly generate DRW parameters.

    Args:
        amp_range (object): An array containing the range of DRW amplitude to simulate.
        log_tau_range (object): An array containing the range of DRW timescale
            (in natural log) to simulate.
        size (int, optional): The number of the set of DRW parameters to generate.
            Defaults to 1.

    Returns:
        array(float): A ndarray of DRW parameters in natural log.
    """

    # validate inputs
    assert (
        len(amp_range) == len(log_tau_range) == 2
    ), "Dimension mismatch: Check the input ranges!"
    assert (np.array(amp_range) > 0).all(), "Amplitudes must be positive!"

    amp_scale = amp_range[1] - amp_range[0]
    log_tau_scale = log_tau_range[1] - log_tau_range[0]
    assert amp_scale > 0, "Lower bound must be smaller than upper bound!"
    assert log_tau_scale > 0, "Lower bound must be smaller than upper bound!"

    # draw samples
    init_tau = np.exp(np.random.rand(size, 1) * log_tau_scale + log_tau_range[0])
    init_amp = np.random.rand(size, 1) * amp_scale + amp_range[0]
    log_drw_param = np.log(np.hstack((init_amp, init_tau)))

    if size == 1:
        return log_drw_param[0]
    else:
        return log_drw_param


def dho_log_param_init(ar_range=[-6, 10], ma_range=[-10, 2], size=1):
    """
    Randomly generate DHO coefficients in the space of the factored polynomials

    The default ranges are optimized for normalized light curves (with a standard
    deviation of unity).

    Args:
        ar_range(object, optional): The range (in natural log) for DHO AR parameters.
            Defaults to [-6, 10].
        ma_range(object, optional): The range (in natural log) for DHO MA parameters.
            Defaults to [-10, 2].
        size (int, optional): The number of the set of DHO parameters to generate.
            Defaults to 1.

    Returns:
        array(float): A ndarray of DHO parameters in natural log.
    """

    ## validate input
    assert (
        len(ar_range) == len(ma_range) == 2
    ), "Dimension mismatch: Check the input ranges!"

    ar_scale = ar_range[1] - ar_range[0]
    ma_scale = ma_range[1] - ma_range[0]
    assert ar_scale > 0, "Lower bound must be smaller than upper bound!"
    assert ma_scale > 0, "Lower bound must be smaller than upper bound!"

    ## draw samples
    p, q = (2, 1)
    log_params = np.random.rand(size, 4)

    # adjust ranges
    log_params[:, :p] = log_params[:, :p] * ar_scale + ar_range[0]
    log_params[:, p:-1] = log_params[:, p:-1] * ma_scale + ma_range[0]

    if size == 1:
        return log_params[0]
    else:
        return log_params


def carma_log_fcoeff_init(
    p, q, ar_range=[-8, 8], ma_range=[-10, 6], ma_mult_range=[-10, 0], size=1
):
    """
    Randomly generate CARMA coefficients in the space of the factored polynomials

    The default ranges are optimized for normalized light curves (with a standard
    deviation of unity).

    Args:
        p (int): The p order of a CARMA(p, q) model.
        q (int): The q order of a CARMA(p, q) model.
        ar_range(object, optional): The range (in natural log) for AR polynomial
            coefficients. Defaults to [-8, 8].
        ma_range(object, optional): The range (in natural log) for MA polynomial
            coefficients. Defaults to [-10, 6].
        ma_mult_range(object, optional): The range for the MA multiplier (the coefficient
            of the highest-order term in the MA characteristic polynomial). Defaults to
            [-10, 0].
        size (int, optional): The number of the set of coefficients to generate.
            Defaults to 1.

    Returns:
        array(float): A ndarray of coeffs for the factored polynomials in natural log.

    .. note:: The notation (index) in the returned coefficients follows that in
        Jones et al. (1981). The last coefficient in the returned array is not part of
        the coefficients, rather a simple multiplying factor of the entire polynomial,
        which is needed to obtain the nominal CARMA representation.
    """
    ## validate inputs
    assert (
        len(ar_range) == len(ma_range) == len(ma_mult_range) == 2
    ), "Dimension mismatch: Check the input ranges!"

    ar_scale = ar_range[1] - ar_range[0]
    ma_scale = ma_range[1] - ma_range[0]
    ma_mult_scale = ma_mult_range[1] - ma_mult_range[0]
    assert ar_scale > 0, "Lower bound must be smaller than upper bound!"
    assert ma_scale > 0, "Lower bound must be smaller than upper bound!"
    assert ma_mult_scale > 0, "Lower bound must be smaller than upper bound!"

    ## draw samples
    dim = int(p + q + 1)
    log_fcoeffs = np.random.rand(size, int(dim))

    # set AR, MA range
    log_fcoeffs[:, :p] = log_fcoeffs[:, :p] * ar_scale + ar_range[0]
    log_fcoeffs[:, p:-1] = log_fcoeffs[:, p:-1] * ma_scale + ma_range[0]

    # set ma_mult range
    log_fcoeffs[:, -1] = log_fcoeffs[:, -1] * ma_mult_scale + ma_mult_range[0]

    if size == 1:
        return log_fcoeffs[0]
    else:
        return log_fcoeffs


def sample_carma(p, q):
    """
    Randomly generate a stationary CARMA model given the orders (p and q).

    Args:
        p (int): The p order of a CARMA(p, q) model.
        q (int): The q order of a CARMA(p, q) model.

    Returns:
        AR and MA coefficients in two separate arrays.
    """
    init_log_fcoeffs = carma_log_fcoeff_init(p, q)
    logAR, logMA = CARMA_term.fcoeffs2carma_log(init_log_fcoeffs, p)
    return np.exp(logAR), np.exp(logMA)


def scipy_opt(
    y,
    gp,
    init_func,
    neg_lp_func,
    n_opt,
    mode="fcoeff",
    debug=False,
    opt_kwargs={},
    opt_options={},
):
    """
    A wrapper for scipy.optimize.minimize method.

    Args:
        y (array(float)): y values of the input time series.
        gp (object): celerite GP object with a proper CARMA kernel.
        init_func (object): A user-provided function to generate initial
            guesses for the optimizer. Defaults to None.
        neg_lp_func (object): A user-provided function to compute negative
            probability given an array of parameters, an array of time series values and
            a celerite GP instance. Defaults to None.
        n_opt (int): Number of iterations to run the optimizer.
        mode (str, optional): The parameter space in which to make proposals, this
            should be determined in the "_fit" functions based on the value of the p
            order. Defaults to "fcoeff".
        debug (bool, optional): Turn on/off debug mode. Defaults to False.
        opt_kwargs (dict, optional): Keyword arguments for scipy.optimize.minimize.
            Defaults to {}.
        opt_options (dict, optional): "options" argument for scipy.optimize.minimize.
            Defaults to {}.

    Returns:
        Best-fit parameters if "debug" is False, an array of scipy.optimize.OptimizeResult objects otherwise.
    """

    initial_params = init_func(size=n_opt)
    dim = gp.kernel.p + gp.kernel.q + 1

    rs = []
    for i in range(n_opt):
        r = minimize(
            neg_lp_func,
            initial_params[i],
            args=(y, gp),
            **opt_kwargs,
            options=opt_options,
        )

        rs.append(r)

    if debug:
        return rs
    else:
        good_rs = [r for r in rs if r.success and r.fun != -np.inf]

        if len(good_rs) == 0:
            return [np.nan] * dim
        else:
            lls = [-r.fun for r in good_rs]
            log_sols = [r.x for r in good_rs]
            best_sol = log_sols[np.argmax(lls)]

            if mode == "fcoeff":
                return np.concatenate(CARMA_term.fcoeffs2carma(best_sol, gp.kernel.p))
            else:
                return np.exp(best_sol)


def drw_fit(
    t,
    y,
    yerr,
    init_func=None,
    neg_lp_func=None,
    optimizer_func=None,
    n_opt=10,
    user_bounds=None,
    scipy_opt_kwargs={},
    scipy_opt_options={},
    debug=False,
):
    """
    Fit DRW.

    Args:
        t (array(float)): Time stamps of the input time series (the default unit is day).
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        init_func (object, optional): A user-provided function to generate initial
            guesses for the optimizer. Defaults to None.
        neg_lp_func (object, optional): A user-provided function to compute negative
            probability given an array of parameters, an array of time series values and
            a celerite GP instance. Defaults to None.
        optimizer_func (object, optional): A user-provided optimizer function.
            Defaults to None.
        n_opt (int, optional): Number of optimizers to run. Defaults to 10.
        user_bounds (list, optional): Parameter boundaries for the default optimizer.
            Defaults to None.
        scipy_opt_kwargs (dict, optional): Keyword arguments for scipy.optimize.minimize.
            Defaults to {}.
        scipy_opt_options (dict, optional): "options" argument for scipy.optimize.minimize.
            Defaults to {}.
        debug (bool, optional): Turn on/off debug mode. Defaults to False.

    Returns:
        array(float): Best-fit DRW parameters
    """
    # re-position lc; compute some stat
    t = t - t[0]
    y = y - np.median(y)
    std = np.std(y)
    min_dt = np.min(t[1:] - t[:-1])

    # init bounds for fitting
    if user_bounds is not None and (len(user_bounds) == 2):
        bounds = user_bounds
    else:
        bounds = [
            (np.log(std / 50), np.log(3 * std)),
            (np.log(min_dt / 5), np.log(t[-1])),
        ]

    # determine negative log probability function
    if neg_lp_func is None:
        neg_lp = partial(neg_lp_flat, bounds=np.array(bounds), mode="param")
    else:
        neg_lp = neg_lp_func

    # define ranges to generate initial guesses
    amp_range = [std / 50, 3 * std]
    log_tau_range = [np.log(min_dt / 5), np.log(t[-1] / 10)]

    # initialize parameter and kernel
    kernel = DRW_term(*drw_log_param_init(amp_range, log_tau_range))
    gp = GP(kernel, mean=0)
    gp.compute(t, yerr)

    # determine initialize function
    if init_func is None:
        init = partial(drw_log_param_init, amp_range, log_tau_range)
    else:
        init = init_func

    # determine optimizer function
    if optimizer_func is None:
        scipy_opt_kwargs.update({"method": "L-BFGS-B", "bounds": bounds})
        opt = partial(
            scipy_opt,
            mode="param",
            opt_kwargs=scipy_opt_kwargs,
            opt_options=scipy_opt_options,
            debug=debug,
        )
    else:
        opt = optimizer_func

    best_fit_return = opt(y, gp, init, neg_lp, n_opt)

    return best_fit_return


def dho_fit(
    t,
    y,
    yerr,
    init_func=None,
    neg_lp_func=None,
    optimizer_func=None,
    n_opt=20,
    user_bounds=None,
    scipy_opt_kwargs={},
    scipy_opt_options={},
    debug=False,
):
    """
    Fit DHO to time series

    The default settings are optimized for normalized LCs.

    Args:
        t (array(float)): Time stamps of the input time series (the default unit is day).
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        init_func (object, optional): A user-provided function to generate initial
            guesses for the optimizer. Defaults to None.
        neg_lp_func (object, optional): A user-provided function to compute negative
            probability given an array of parameters, an array of time series values and
            a celerite GP instance. Defaults to None.
        optimizer_func (object, optional): A user-provided optimizer function.
            Defaults to None.
        n_opt (int, optional): Number of optimizers to run.. Defaults to 20.
        user_bounds (list, optional): Parameter boundaries for the default optimizer and
            the default flat prior. Defaults to None.
        scipy_opt_kwargs (dict, optional): Keyword arguments for scipy.optimize.minimize.
            Defaults to {}.
        scipy_opt_options (dict, optional): "options" argument for scipy.optimize.minimize.
            Defaults to {}.
        debug (bool, optional): Turn on/off debug mode. Defaults to False.

    Raises:
        celerite.solver.LinAlgError: For non-positive definite autocovariance matrices.

    Returns:
        array(float): Best-fit DHO parameters
    """

    # determine user defined boundaries if any
    if user_bounds is not None and (len(user_bounds) == 4):
        bounds = user_bounds
    else:
        bounds = [(-15, 15)] * 4
        bounds[2:] = [(a[0] - 8, a[1] - 8) for a in bounds[2:]]

    # re-position/normalize lc
    t = t - t[0]
    y = y - np.median(y)
    y_std = mad(y) * 1.4826
    y = y / y_std
    yerr = yerr / y_std

    # determine negative log probability function
    if neg_lp_func is None:
        neg_lp = partial(neg_lp_flat, bounds=np.array(bounds), mode="param")
    else:
        neg_lp = neg_lp_func

    # initialize parameter, kernel and GP
    kernel = DHO_term(*dho_log_param_init())
    gp = GP(kernel, mean=0)
    gp.compute(t, yerr)

    # determine initialize function
    if init_func is None:
        init = partial(dho_log_param_init)
    else:
        init = init_func

    # determine the optimizer function
    if optimizer_func is None:
        scipy_opt_kwargs.update({"method": "L-BFGS-B", "bounds": bounds})
        opt = partial(
            scipy_opt,
            mode="param",
            opt_kwargs=scipy_opt_kwargs,
            opt_options=scipy_opt_options,
            debug=debug,
        )
    else:
        opt = optimizer_func

    # get best-fit solution & adjust MA params (multiply by y_std)
    # if not fit found, return all nan
    best_fit_return = opt(y, gp, init, neg_lp, n_opt)
    try:
        best_fit_return[2:] = best_fit_return[2:] * y_std
    except:
        pass

    return best_fit_return


def carma_fit(
    t,
    y,
    yerr,
    p,
    q,
    init_func=None,
    neg_lp_func=None,
    optimizer_func=None,
    n_opt=20,
    user_bounds=None,
    scipy_opt_kwargs={},
    scipy_opt_options={},
    debug=False,
):
    """
    Fit an arbitrary CARMA model

    The default settings are optimized for normalized LCs.

    Args:
        t (array(float)): Time stamps of the input time series (the default unit is day).
        y (array(float)): y values of the input time series.
        yerr (array(float)): Measurement errors for y values.
        p (int): The p order of a CARMA(p, q) model.
        q (int): The q order of a CARMA(p, q) model.
        init_func (object, optional): A user-provided function to generate initial
            guesses for the optimizer. Defaults to None.
        neg_lp_func (object, optional): A user-provided function to compute negative
            probability given an array of parameters, an array of time series values and
            a celerite GP instance. Defaults to None.
        optimizer_func (object, optional): A user-provided optimizer function.
            Defaults to None.
        n_opt (int, optional): Number of optimizers to run.
            Defaults to 20.
        user_bounds (array(float), optional): Parameter boundaries for the default
            optimizer. If p > 2, these are boundaries for the coefficients of the
            factored polynomial. Defaults to None.
        scipy_opt_kwargs (dict, optional): Keyword arguments for scipy.optimize.minimize.
            Defaults to {}.
        scipy_opt_options (dict, optional): "options" argument for scipy.optimize.minimize.
            Defaults to {}.
        debug (bool, optional): Turn on/off debug mode. Defaults to False.

    Raises:
        celerite.solver.LinAlgError: For non-positive definite autocovariance matrices.

    Returns:
        array(float): Best-fit parameters
    """
    # set core config
    dim = int(p + q + 1)
    mode = "fcoeff" if p > 2 else "param"

    # init bounds for fitting
    if user_bounds is not None and (len(user_bounds) == dim):
        bounds = user_bounds
    else:
        bounds = [(-15, 15)] * dim
        bounds[p:-1] = [(a[0] - 5, a[1] - 5) for a in bounds[p:-1]]
        bounds[-1] = (-15, 5)

    # re-position lc
    t = t - t[0]
    y = y - np.median(y)
    y_std = mad(y) * 1.4826
    y = y / y_std
    yerr = yerr / y_std

    # initialize parameter and kernel
    ARpars, MApars = sample_carma(p, q)
    kernel = CARMA_term(np.log(ARpars), np.log(MApars))
    gp = GP(kernel, mean=0)
    gp.compute(t, yerr)

    # determine/set init func
    if init_func is not None:
        init = init_func
    else:
        init = partial(carma_log_fcoeff_init, p, q)

    # determine/set negative log probability function
    if neg_lp_func is None:
        neg_lp = partial(neg_lp_flat, bounds=np.array(bounds), mode=mode)
    else:
        neg_lp = neg_lp_func

    # determine/set optimizer function
    if optimizer_func is None:
        scipy_opt_kwargs.update({"method": "L-BFGS-B", "bounds": bounds})
        opt = partial(
            scipy_opt,
            mode=mode,
            opt_kwargs=scipy_opt_kwargs,
            opt_options=scipy_opt_options,
            debug=debug,
        )
    else:
        opt = optimizer_func

    # get best-fit solution & adjust MA params (multiply by y_std)
    # if not fit found, return all nan
    best_fit_return = opt(y, gp, init, neg_lp, n_opt)
    try:
        best_fit_return[p:] = best_fit_return[p:] * y_std
    except:
        pass

    return best_fit_return


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['eztao', 'eztao.carma', 'eztao.ts', 'eztao.viz']

package_data = \
{'': ['*']}

install_requires = \
['celerite>=0.3.0',
 'emcee>=3.0.0',
 'importlib-metadata>=2.0.0',
 'matplotlib>=3.3.0,<4.0.0',
 'numba>=0.51.0',
 'scipy>=1.5']

setup_kwargs = {
    'name': 'eztao',
    'version': '0.4.1',
    'description': 'A toolkit for Active Galactic Nuclei (AGN) time-series analysis.',
    'long_description': '![tests](https://github.com/ywx649999311/EzTao/workflows/tests/badge.svg)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/ywx649999311/EzTao/v0.4.1?filepath=docs/notebooks)\n[![Documentation Status](https://readthedocs.org/projects/eztao/badge/?version=latest)](https://eztao.readthedocs.io/en/latest/)\n<a href="https://ascl.net/2201.001"><img src="https://img.shields.io/badge/ascl-2201.001-blue.svg?colorB=262255" alt="ascl:2201.001" /></a>\n# EzTao (易道)\n**EzTao** is a toolkit for conducting AGN time-series/variability analysis, mainly utilizing the continuous-time auto-regressive moving average model (CARMA)\n\n## Installation\n<!--- ```\npip install eztao\n```\nor (from master)--->\n```\npip install git+https://github.com/ywx649999311/EzTao.git\n```\n### Dependencies\n>```\n>python = "^3.7"\n>celerite = ">= 0.3.0"\n>matplotlib = "^3.3.0"\n>scipy = "> 1.5.0"\n>numba = ">= 0.51.0"\n>emcee = ">=3.0.0"\n>```\n\n## Quick Examples\nLet\'s first simulate a DRW/CARMA(1,0) process with a variance of 0.3^2 and a relaxation timescale of 100 days. This time series will have a total of 200 data points and span 10 years.\n```python\nimport numpy as np\nimport matplotlib.pyplot as plt\nfrom eztao.carma import DRW_term\nfrom eztao.ts import gpSimRand\n\n# define a DRW kernel & and simulate a process\namp = 0.2\ntau = 100\nDRW_kernel = DRW_term(np.log(amp), np.log(tau))\nt, y, yerr = gpSimRand(DRW_kernel, 10, 365*10, 200)\n\n# now, plot it\nfig, ax = plt.subplots(1,1, dpi=150, figsize=(8,3))\nax.errorbar(t, y, yerr, fmt=\'.\')\n...\n```\n![drw_sim](include/drw_sim.png)\n\nWe can fit the simulated time series to the DRW model and see how well we can recover the input parameters.\n```python\nfrom eztao.ts import drw_fit\n\nbest_fit = drw_fit(t, y, yerr)\nprint(f\'Best-fit DRW parameters: {best_fit}\')\n```\n```shell\nBest-fit DRW parameters: [0.17356983 88.36262467]\n```\n\nHow does the power spectrum density (PSD) compare?\n```python\nfrom eztao.carma import gp_psd\n\n# get psd functions\ntrue_psd = gp_psd(DRW_kernel)\nbest_psd = gp_psd(DRW_term(*np.log(best_fit)))\n\n# plot\nfig, ax = plt.subplots(1,1, dpi=150, figsize=(6,3))\nfreq = np.logspace(-5, 2)\nax.plot(freq, true_psd(freq), label=\'Input PSD\')\nax.plot(freq, best_psd(freq), label=\'Best-fit PSD\')\n...\n```\n![drw_psd](include/drw_psd.png)\n\n__Note:__ How well the input and best-fit PSD match is up to how good the best-fit parameters are, which is highly influenced by the quality of the input time series. \n\nFor more examples, please check out the [online documentation](https://eztao.readthedocs.io/en/latest/) or run the tutorial notebooks at ->\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/ywx649999311/EzTao/v0.4.0?filepath=docs/notebooks).\n\n## Development\n`poetry` is used to solve dependencies and to build/publish this package. Below shows how setup the environment for development (assuming you already have `poetry` installed on your machine). _**Warning:**_ `poetry` is having issue installing `llvmlite = 0.34.0` (used for `eztao = ^0.4.0`) under Python 3.9. The issue disappears for Python 3.8.\n\n1. Clone this repository, and enter the repository folder.\n2. Create a python virtual environment and activate it (the virtual environment name must be \'.venv\'). \n    ```\n    python -m venv .venv\n    source .venv/bin/activate\n    ```\n3. Install dependencies and **EzTao** in editable mode.\n   ```\n   poetry install\n   ```\n\nNow you should be ready to start adding new features. Be sure to checkout the normal practice regarding how to use `poetry` on its website. When you are ready to push, also make sure the poetry.lock file is checked-in if any dependency has changed. \n\n## Citation\nWe are working on a paper to describe the full implementation of **EzTao**. In the meantime, if you find **EzTao** useful for your research, please consider acknowledging **EzTao** using the following:\n\n```\n@MISC{Yu2022,\n       author = {{Yu}, Weixiang and {Richards}, Gordon T.},\n        title = "{EzTao: Easier CARMA Modeling}",\n     keywords = {Software},\n howpublished = {Astrophysics Source Code Library, record ascl:2201.001},\n         year = 2022,\n        month = jan,\n          eid = {ascl:2201.001},\n        pages = {ascl:2201.001},\narchivePrefix = {ascl},\n       eprint = {2201.001},\n       adsurl = {https://ui.adsabs.harvard.edu/abs/2022ascl.soft01001Y},\n      adsnote = {Provided by the SAO/NASA Astrophysics Data System}\n}\n```\n\n',
    'author': 'Weixiang Yu',
    'author_email': 'wy73@drexel.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ywx649999311/EzTao',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
A collection of GP kernels that express the autovariance structure of CARMA models using 
celerite.
"""

import numpy as np
from celerite import terms
from numba import njit, float64, complex128, int32, vectorize
import warnings

__all__ = ["acf", "DRW_term", "DHO_term", "CARMA_term"]


@njit(complex128[:](complex128[:]))
def _compute_roots(coeffs):
    """Internal jitted function to compute roots"""

    # find roots using np and make roots that are almost real real
    roots = np.roots(coeffs)
    roots[np.abs(roots.imag) < 1e-10] = roots[np.abs(roots.imag) < 1e-10].real
    roots = roots[roots.real.argsort()]  # ascending sort by real part

    return roots


@njit(float64[:](float64[:]))
def _compute_exp(params):
    """Internal jitted np.exp"""
    return np.exp(params)


@njit(float64[:](float64[:], float64[:]))
def polymul(poly1, poly2):
    """Multiply two polynomials

    Inputs are the coefficients of the polynomials ranked by order (high
    to low). Internally, this function operates from low order to high order.
    """
    ## convert from high->low to low->high
    poly1 = poly1[::-1]
    poly2 = poly2[::-1]
    poly1_len = poly1.shape[0]
    poly2_len = poly2.shape[0]
    c = np.zeros(poly1_len + poly2_len - 1)

    for i in np.arange(poly1_len):
        for j in np.arange(poly2_len):
            c[i + j] += poly1[i] * poly2[j]

    ## convert back to high->low
    return c[::-1]


@njit(float64[:](float64[:]))
def fcoeffs2coeffs(fcoeffs):
    """Convert coeffs of a factored polynomial to the coeffs of the product

    The input fcoeffs follow the notation (index) in Jones et al. (1981) with the last
    element being an additional multiplying factor (the coeff of the highest-order term
    in the final expanded polynomial).

    :meta private:
    """
    size = fcoeffs.shape[0] - 1
    odd = bool(size & 0x1)
    nPair = size // 2
    poly = fcoeffs[-1:]  # The coeff of highest order term in the product

    if odd:
        poly = polymul(poly, np.array([1.0, fcoeffs[-2]]))

    for p in np.arange(nPair):
        poly = polymul(poly, np.array([1.0, fcoeffs[p * 2], fcoeffs[p * 2 + 1]]))

    # the returned is high->low
    return poly


def _roots2coeffs(roots):
    """Generate factored polynomial from roots

    The notation (index) for the factored polynomial follows that in
    Jones et al. (1981). Note: No multiplying is returned.
    """
    coeffs = []
    size = len(roots)
    odd = bool(size & 0x1)
    rootsComp = roots[roots.imag != 0]
    rootsReal = roots[roots.imag == 0]
    nCompPair = len(rootsComp) // 2
    nRealPair = len(rootsReal) // 2

    for i in range(nCompPair):
        root1 = rootsComp[i]
        root2 = rootsComp[i + 1]
        coeffs.append(-(root1.real + root2.real))
        coeffs.append((root1 * root2).real)

    for i in range(nRealPair):
        root1 = rootsReal[i]
        root2 = rootsReal[i + 1]
        coeffs.append(-(root1.real + root2.real))
        coeffs.append((root1 * root2).real)

    if odd:
        coeffs.append(-rootsReal[-1].real)
    return coeffs


@njit(complex128[:](complex128[:], float64[:], float64[:]))
def acf(arroots, arparam, maparam):
    """Get ACVF coefficients given CARMA parameters

    The CARMA noation (index) folows that in Brockwell et al. (2001).

    Args:
        arroots (array(complex)): AR roots in a numpy array
        arparam (array(float)): AR parameters in a numpy array
        maparam (array(float)): MA parameters in a numpy array

    Returns:
        array(complex): ACVF coefficients, each element correspond to a root.
    """
    p = arparam.shape[0]
    q = maparam.shape[0] - 1
    sigma = maparam[0]

    # MA param into Kelly's notation
    # arparam = np.array(arparam)
    maparam = np.array([x / sigma for x in maparam])

    # init acf product terms
    num_left = np.zeros(p, dtype=np.complex128)
    num_right = np.zeros(p, dtype=np.complex128)
    denom = -2 * arroots.real + np.zeros_like(arroots) * 1j

    for k in range(q + 1):
        num_left += maparam[k] * np.power(arroots, k)
        num_right += maparam[k] * np.power(np.negative(arroots), k)

    for j in range(1, p):
        root_idx = np.arange(p)
        root_k = arroots[np.roll(root_idx, j)]
        denom *= (root_k - arroots) * (np.conj(root_k) + arroots)

    return sigma ** 2 * num_left * num_right / denom


class DRW_term(terms.Term):
    r"""
    Damped Random Walk (DRW) term (kernel)

    .. math::

        k_{DRW}(\Delta t) = \sigma^2\,e^{-\Delta t/\tau}

    with the parameters ``log_amp`` and ``log_tau``.

    Args:
        log_amp (float): The natural log of the RMS amplitude of the DRW process.
        log_tau (float): The natural log of the characteristic timescale of the DRW
            process.

    .. note::
        Conversions between EzTao DRW parameters and some other DRW representations
        seen in the literature:

        .. math::

            \mathrm{SF_{\infty}} = 2*\sigma^2

        .. math::

            \sigma^2 = \tau*\sigma_{KBS}^2/2

        .. math::

            \tau = 1/\alpha_1; \,\sigma_{KBS} = \beta_0

        see MacLeod et al. (2010) for SF_{\infty}} and Kelly et al. (2009)
        for \sigma_{KBS}. \alpha_1 and \beta_0 are the AR and MA parameters for a
        CARMA(1,0) model, respectively.
    """

    parameter_names = ("log_amp", "log_tau")

    def get_real_coefficients(self, params):
        """Get ``alpha_real`` and ``beta_real`` (coeffs of celerite's real kernel)

        Args:
            params (array(float)): Parameters of this kernel.

        Returns:
            (``alpha_real``, ``beta_real``).
        """
        log_amp, log_tau = params
        return (np.exp(2 * log_amp), 1 / np.exp(log_tau))

    def get_perturb_amp(self):
        """Get the amplitude of the perturbing noise (beta_0) in DRW

        Returns:
            The amplitude of the perturbing noise (beta_0) in the current DRW.
        """
        log_amp, log_tau = self.get_parameter_vector()
        return self.perturb_amp(log_amp, log_tau)

    @staticmethod
    def perturb_amp(log_amp, log_tau):
        """Compute the amplitude of the perturbing noise (beta_0) in DRW.

        Args:
            log_amp (float): The natural log of the RMS amplitude of the DRW process.
            log_tau (float): The natural log of the characteristic timescale of the DRW
                process.

        Returns:
            The amplitude of the perturbing noise (beta_0) in the DRW specified by the
            input parameters.
        """
        return np.exp((2 * log_amp - np.log(1 / 2) - log_tau) / 2)

    def get_rms_amp(self):
        """Get the RMS amplitude of this DRW process.

        Returns:
            The RMS amplitude of this DRW process.
        """
        log_amp, log_tau = self.get_parameter_vector()
        return np.exp(log_amp)

    def get_carma_parameter(self):
        """Get DRW parameters in CARMA notation (alpha_*/beta_*).

        Returns:
            [alpha_1, beta_0].
        """
        return [1 / np.exp(self.get_parameter("log_tau")), self.get_perturb_amp()]

    @property
    def p(self):
        return 1

    @property
    def q(self):
        return 0


class CARMA_term(terms.Term):
    """A general-purpose CARMA term (kernel)

    Args:
        log_arpars (array(float)): Natural log of the AR coefficients.
        log_mapars (array(float)): Natural log of the MA coefficients.
    """

    def __init__(self, log_arpars, log_mapars, *args, **kwargs):

        arpar_temp = "log_a{}"
        mapar_temp = "log_b{}"
        arpar_names = ("log_a1",)
        mapar_names = ("log_b0",)

        # set order & trigger roots/acf computation
        log_pars = np.append(log_arpars, log_mapars)
        self._p = len(log_arpars)
        self._q = len(log_mapars) - 1
        self._dim = self._p + self._q + 1
        self._compute(log_pars)

        # check if stationary
        self._arroots = _compute_roots(np.append([1 + 0j], self._pars[: self._p]))
        if (self._arroots.real > 0).any():
            print("Warning: CARMA process is not stationary!")

        # loop over par array to find out how many params
        for i in range(2, self._p + 1):
            arpar_names += (arpar_temp.format(i),)

        for i in range(1, self._q + 1):
            mapar_names += (mapar_temp.format(i),)

        self.parameter_names = arpar_names + mapar_names
        super().__init__(*log_pars, **kwargs)

    @property
    def p(self):
        return self._p

    @property
    def q(self):
        return self._q

    def _compute(self, params):
        """Compute important CARMA parameters"""
        self._pars = _compute_exp(params)
        self._arroots = _compute_roots(np.append([1 + 0j], self._pars[: self._p]))
        self.acf = acf(self._arroots, self._pars[: self._p], self._pars[self._p :])
        self.mask = self._arroots.imag != 0

    def set_log_fcoeffs(self, log_fcoeffs):
        """Set kernel parameters

        Use coeffs of the factored polynomial to set CARMA paramters, note that the
        last input coeff is always the coeff for the highest-order MA differential.
        While performing the conversion, 1 is added to the AR coeffs to maintain the
        same formatting (AR polynomials always have the highest order coeff to be 1).

        Args:
            log_fcoeffs (array(float)): Natural log of the coefficients for the
                factored characteristic polynomial, with the last coeff being an
                additional multiplying factor on this polynomial.

        """
        if log_fcoeffs.shape[0] != (self._dim):
            raise ValueError("Dimension mismatch!")

        fcoeffs = _compute_exp(log_fcoeffs)
        ARpars = fcoeffs2coeffs(np.append(fcoeffs[: self._p], [1]))[1:]
        MApars = fcoeffs2coeffs(fcoeffs[self._p :])[::-1]
        self.set_parameter_vector(np.log(np.append(ARpars, MApars)))

    def get_real_coefficients(self, params):
        """
        Get arrays of ``alpha_real`` and ``beta_real`` (coefficients of celerite's
        real kernel)

        Args:
            params (array(float)): Parameters of this kernel.

        Returns:
            Arrays of ``alpha_real`` and ``beta_real``, one for each.
        """

        # trigger re_compute & get celerite coeffs
        self._compute(params)
        acf_real = self.acf[~self.mask]
        roots_real = self._arroots[~self.mask]

        ar = acf_real[:].real
        cr = -roots_real[:].real

        return (ar, cr)

    def get_complex_coefficients(self, params):
        """
        Get arrays of ``alpha_complex_real``, ``alpha_complex_imag``,
        ``beta_complex_real`` and ``beta_complex_imag`` (coefficients of celerite's
        complex kernel)

        Args:
            params (array(float)): Parameters of this kernel.

        Returns:
            Arrays of ``alpha_complex_real``, ``alpha_complex_imag``,
            ``beta_complex_real`` and ``beta_complex_imag``, one for each.
        """

        acf_complex = self.acf[self.mask]
        roots_complex = self._arroots[self.mask]

        ac = 2 * acf_complex[::2].real
        bc = 2 * acf_complex[::2].imag
        cc = -roots_complex[::2].real
        dc = -roots_complex[::2].imag

        return (ac, bc, cc, dc)

    def get_rms_amp(self):
        """Get the RMS amplitude of this CARMA kernel

        Returns:
            The RMS amplitude of this CARMA kernel.
        """
        log_pars = self.get_parameter_vector()
        return self.rms_amp(log_pars[: self.p], log_pars[self.p :])

    @staticmethod
    def rms_amp(log_arpars, log_mapars):
        """Compute the RMS amplitude of a CARMA kernel

        Args:
            log_arpars (array(float)): Natural log of the AR coefficients.
            log_mapars (array(float)): Natural log of the MA coefficients.

        Returns:
            The RMS amplitude of the CARMA kernel specified by the input parameters.
        """
        _p = len(log_arpars)
        _pars = _compute_exp(np.append(log_arpars, log_mapars))
        _arroots = _compute_roots(np.append([1 + 0j], _pars[:_p]))
        _acf = acf(_arroots, _pars[:_p], _pars[_p:])

        return np.sqrt(np.abs(np.sum(_acf)))

    @staticmethod
    def carma2fcoeffs_log(log_arpars, log_mapars):
        """Get the representation of a CARMA kernel in the factored polynomial space

        Args:
            log_arpars (array(float)): Natural log of the AR coefficients.
            log_mapars (array(float)): Natural log of the MA coefficients.

        Returns:
            array(float): The coefficients (in natural log) of the factored polymoical
                for the CARMA kernel specified by the input parameters. The last coeff
                is a multiplying factor of the returned polynomial.
        """

        _p = len(log_arpars)
        _q = len(log_mapars) - 1
        _pars = _compute_exp(np.append(log_arpars, log_mapars))
        _arroots = _compute_roots(np.append([1 + 0j], _pars[:_p]))
        _maroots = _compute_roots(np.array(_pars[_p:][::-1], dtype=np.complex128))
        ma_mult = _pars[-1:]  ## the multiplying factor
        ar_coeffs = _roots2coeffs(_arroots)

        if _q > 0:
            ma_coeffs = _roots2coeffs(_maroots)
            ma_coeffs.append(ma_mult[0])
        else:
            ma_coeffs = ma_mult

        return np.log(np.append(ar_coeffs, ma_coeffs))

    @staticmethod
    def fcoeffs2carma_log(log_fcoeffs, p):
        """Get the representation of a CARMA kernel in the nominal CARMA parameter space

        Args:
            log_coeffs (array(float)): The array of coefficients for the factored
                polynomial with the last coeff being a multiplying factor of the
                polynomial.
            p (int): The p order of the CARMA kernel.

        Returns:
            Natural log of the AR and MA parameters in two separate arrays.
        """

        fcoeffs = np.exp(log_fcoeffs)

        # Append one to AR fcoeffs as the multiplying factor; MA foceffs has that included.
        # Index in CARMA for AR: high -> low; for MA: low -> high
        ARpars = fcoeffs2coeffs(np.append(fcoeffs[:p], [1]))[1:]
        MApars = fcoeffs2coeffs(fcoeffs[p:])[::-1]

        return np.log(ARpars), np.log(MApars)

    @staticmethod
    def carma2fcoeffs(log_arpars, log_mapars):
        """Get the representation of a CARMA kernel in the factored polynomial space

        A wrapper of `CARMA_term.carma2fcoeffs_log` for backward compatibility. This
        function will be deprecated in future releases.
        """
        warnings.warn("Use carma2fcoeffs_log instead", DeprecationWarning)
        log_fcoeffs = CARMA_term.carma2fcoeffs_log(log_arpars, log_mapars)
        return np.exp(log_fcoeffs)

    @staticmethod
    def fcoeffs2carma(log_fcoeffs, p):
        """Get the representation of a CARMA kernel in the nominal CARMA parameter space

        A wrapper of `CARMA_term.fcoeffs2carma_log` for backward compatibility. This
        function will be deprecated in future releases.
        """
        warnings.warn("Use fcoeffs2carma_log instead", DeprecationWarning)
        log_ar, log_ma = CARMA_term.fcoeffs2carma_log(log_fcoeffs, p)
        return np.exp(log_ar), np.exp(log_ma)


class DHO_term(CARMA_term):
    """Damped Harmonic Oscillator (DHO) term (kernel)

    Args:
        log_a1 (float): Natural log of the DHO parameter a1.
        log_a2 (float): Natural log of the DHO parameter a2.
        log_b0 (float): Natural log of the DHO parameter b0.
        log_b1 (float): Natural log of the DHO parameter b1.
    """

    def __init__(self, log_a1, log_a2, log_b0, log_b1, *args, **kwargs):
        """Initiate the DHO term."""
        super(DHO_term, self).__init__([log_a1, log_a2], [log_b0, log_b1], **kwargs)
